use crate::ast::named::Named;
use crate::ast::*;
use crate::name_analysis::fqn_context::FqnContext;
use crate::name_analysis::symbol::{FunctionSymbol, ModuleSymbol, SourceDefinition, Symbol, VariableSymbol};
use crate::name_analysis::symbol_table::{SymbolInsertError, SymbolTable};
use crate::name_analysis::DiagnosticsContainer;
use codespan_reporting::diagnostic::{Diagnostic, Label};
use std::range::Range;

fn handle_insert_error(
    err: SymbolInsertError,
    error_symbol_name: &str,
    error_file_id: usize,
    error_range: Range<usize>,
    symbol_types: &str,
    diagnostics: &mut DiagnosticsContainer,
) {
    match err {
        SymbolInsertError::SymbolAlreadyDefined(s) => {
            let already_defined_definition = s.definition();

            diagnostics.add(
                Diagnostic::error()
                    .with_message(format!(
                        "{} symbol '{}' already defined in the current scope.",
                        symbol_types, error_symbol_name,
                    ))
                    .with_label(
                        Label::primary(error_file_id, error_range)
                            .with_message("Symbol duplicated here."),
                    )
                    .with_label(
                        Label::secondary(
                            already_defined_definition.file_id(),
                            already_defined_definition.range(),
                        )
                        .with_message("Symbol defined here."),
                    ),
            );
        }
    }
}

pub(super) fn gather_module_level_declaration(
    declaration: &mut ModuleLevelDeclaration,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    use ModuleLevelDeclaration::*;
    match declaration {
        Module(module_declaration) => {
            gather_module_declaration(module_declaration, symbol_table, fqn_context, diagnostics);
        }
        Function(function_definition) => {
            gather_function_definition(function_definition, symbol_table, fqn_context, diagnostics)
        }
        _ => todo!(),
    }
}

fn gather_module_declaration(
    declaration: &mut ModuleDeclaration,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    // 1. Add mod identifier symbol
    // 2. Update fqn context
    // 3. Push scope
    // 4. Process declarations
    // 5. Pop scope

    let module_name = declaration.identifier.name();

    let insert_result = symbol_table.insert(
        &module_name,
        Symbol::Module(ModuleSymbol::new(
            &fqn_context.resolve(&module_name),
            &module_name,
            declaration.is_public,
            &declaration.identifier,
        )),
    );

    if let Err(err) = insert_result {
        handle_insert_error(
            err,
            &module_name,
            declaration.identifier.file_id,
            declaration.identifier.range,
            "module/type",
            diagnostics,
        )
    }

    fqn_context.push(module_name.to_string());

    symbol_table.push_scope(&format!("Module '{}' Scope", module_name));
    for inner_declaration in &mut declaration.declarations {
        gather_module_level_declaration(inner_declaration, symbol_table, fqn_context, diagnostics);
    }
    symbol_table.pop_scope()
}

fn gather_function_definition(
    function: &mut FunctionDefinition,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    let declared_name = function.identifier.name();
    let resolved_name = fqn_context.resolve(&declared_name);

    let insert_result = symbol_table.insert(
        &declared_name,
        Symbol::Function(FunctionSymbol::new(
            &resolved_name,
            &declared_name,
            function.is_public,
            &function.identifier,
        )),
    );

    if let Err(err) = insert_result {
        handle_insert_error(
            err,
            &declared_name,
            function.identifier.file_id,
            function.identifier.range,
            "function/variable",
            diagnostics,
        )
    }

    function
        .identifier
        .set_scope_id(symbol_table.current_scope_id());

    symbol_table.push_scope(&format!("FunctionScope({})", resolved_name));
    gather_parameters(
        &mut function.parameters,
        symbol_table,
        fqn_context,
        diagnostics,
    );
    gather_function_body(&mut function.body, symbol_table, fqn_context, diagnostics);
    symbol_table.pop_scope();
}

fn gather_parameters(
    parameters: &mut Parameters,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    for parameter in &mut parameters.0 {
        gather_parameter(parameter, symbol_table, fqn_context, diagnostics);
    }
}

fn gather_parameter(
    parameter: &mut Parameter,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    let parameter_name = parameter.identifier.name();

    let insert_result = symbol_table.insert(
        &parameter_name,
        Symbol::Variable(VariableSymbol::new(
            &parameter_name,
            false,
            SourceDefinition::from_identifier(&parameter.identifier),
        )),
    );

    if let Err(err) = insert_result {
        handle_insert_error(
            err,
            &parameter_name,
            parameter.identifier.file_id,
            parameter.identifier.range,
            "function/variable",
            diagnostics,
        )
    }

    parameter
        .identifier
        .set_scope_id(symbol_table.current_scope_id());
}

fn gather_function_body(
    function_body: &mut FunctionBody,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    use crate::ast::FunctionBody::*;
    match function_body {
        Block(block) => gather_block_statement_inner(block, symbol_table, fqn_context, diagnostics),
        _ => todo!(),
    }
}

fn gather_block_statement(
    block: &mut BlockStatement,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    symbol_table.push_scope("BlockStatementScope");
    gather_block_statement_inner(block, symbol_table, fqn_context, diagnostics);
    symbol_table.pop_scope();
}

fn gather_block_statement_inner(
    block: &mut BlockStatement,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    for statement in &mut block.statements {
        gather_statement(statement, symbol_table, fqn_context, diagnostics);
    }
    if let Some(expression) = &mut block.expression {
        gather_expression(expression, symbol_table, diagnostics);
    }
}

fn gather_statement(
    statement: &mut Statement,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    use crate::ast::Statement::*;
    match statement {
        BlockStatement(block) => {
            gather_block_statement(block, symbol_table, fqn_context, diagnostics)
        }
        VariableDeclarationStatement(variable_declaration) => {
            gather_variable_declaration(variable_declaration, symbol_table, diagnostics)
        }
        AssignStatement(assign_statement) => {
            gather_assign_statement(assign_statement, symbol_table, fqn_context, diagnostics)
        }
        _ => todo!(),
    }
}

fn gather_variable_declaration(
    variable_declaration: &mut VariableDeclarationStatement,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    let variable_name = variable_declaration.identifier.name();

    let insert_result = symbol_table.insert(
        &variable_name,
        Symbol::Variable(VariableSymbol::new( 
            &variable_name,
            variable_declaration.is_mutable,
            SourceDefinition::from_identifier(&variable_declaration.identifier),
        )),
    );

    if let Err(err) = insert_result {
        handle_insert_error(
            err,
            &variable_name,
            variable_declaration.identifier.file_id,
            variable_declaration.identifier.range,
            "function/variable",
            diagnostics,
        )
    }

    variable_declaration
        .identifier
        .set_scope_id(symbol_table.current_scope_id());

    if let Some(initializer) = &mut variable_declaration.initializer {
        gather_expression(initializer, symbol_table, diagnostics);
    }
}

fn gather_assign_statement(
    assign_statement: &mut AssignStatement,
    symbol_table: &mut SymbolTable,
    fqn_context: &mut FqnContext,
    diagnostics: &mut DiagnosticsContainer,
) {
    gather_expression(&mut assign_statement.lhs, symbol_table, diagnostics);
    gather_expression(&mut assign_statement.rhs, symbol_table, diagnostics);
}

fn gather_expression(
    expression: &mut Expression,
    symbol_table: &mut SymbolTable,
    diagnostics: &mut DiagnosticsContainer,
) {
    use crate::ast::Expression::*;
    match expression {
        FullyQualifiedName(fully_qualified_name) => {
            gather_fully_qualified_name(fully_qualified_name, symbol_table);
        }
        _ => {}
    }
}

fn gather_fully_qualified_name(
    fully_qualified_name: &mut FullyQualifiedName,
    symbol_table: &mut SymbolTable,
) {
    fully_qualified_name.set_scope_id(symbol_table.current_scope_id());
}
